# Copyright (c) 2017 David Preece, All rights reserved.
# 
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""AWS CloudWatch compatible logger - logs to stdout if not running under AWS"""

import json.decoder
import requests
import boto3
import logging
import json
from botocore.exceptions import ClientError
from multiprocessing import Queue, Process


class LogHandler(logging.Handler):
    def __init__(self, group, stream, blacklist=None):
        try:
            # see if we are logging verbosely
            dynamic_data_text = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document').text
            ssm = boto3.client('ssm', region_name=json.loads(dynamic_data_text)['region'])
            try:
                verbose_string = ssm.get_parameter(Name='/20ft/verbose')['Parameter']['Value']
            except ClientError:  # string not written yet
                verbose_string = 'False'
            level = logging.DEBUG if verbose_string == 'True' or verbose_string == 'true' else logging.INFO

            # find instance ID
            iid = requests.get('http://169.254.169.254/latest/meta-data/instance-id').text
            stream_name = stream + '/' + iid

        except requests.exceptions.ConnectionError:  # not running under AWS
            level = logging.DEBUG
            stream_name = None

        super().__init__(level)

        # wake the actual logger and ensure it has a queue
        self.queue = Queue()  # so log messages can be delivered
        logging.basicConfig(level=level, handlers=[self])

        # delivery process
        self.process = Process(target=self.background, args=(group, stream_name, self.queue))
        self.blacklist = blacklist if blacklist is not None else []
        self.process.start()

    def stop(self):
        self.queue.put(None)
        self.process.join()

    def emit(self, record):
        # urllib tries to log itself doing all sorts
        if record.name.startswith('urllib3') or record.name.startswith('botocore'):
            return
        # otherwise enqueue the record
        self.queue.put(record)

    def background(self, group, stream, queue):
        """Runs as a background process delivering the logs as they arrive (to avoid stalling the event loop)"""
        sequence_token = None
        if stream is not None:  # on aws
            # create a cloud watch log client
            dynamic_data_text = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document').text
            aws_logger = boto3.client('logs', region_name=json.loads(dynamic_data_text)['region'])

            # create the group and stream if need be
            groups = aws_logger.describe_log_groups(
                logGroupNamePrefix=group
            )['logGroups']
            groups_dict = {g['logGroupName']: g for g in groups}
            if group not in groups_dict.keys():
                aws_logger.create_log_group(
                        logGroupName=group
                )
            streams = aws_logger.describe_log_streams(
                    logGroupName=group,
                    logStreamNamePrefix=stream
            )['logStreams']
            streams_dict = {s['logStreamName']: s for s in streams}
            if stream not in streams_dict.keys():
                result = aws_logger.create_log_stream(
                        logGroupName=group,
                        logStreamName=stream
                )
            else:
                try:
                    sequence_token = streams_dict[stream]['uploadSequenceToken']
                except KeyError:
                    pass

        # loop picking logs off the queue and delivering
        formatter = logging.Formatter(fmt='%(levelname)-8s %(message)s')
        while True:
            try:
                record = queue.get()
            except KeyboardInterrupt:  # so we get 'record is None' to close, regardless of if it came off the queue
                record = None
            if record is None:
                return

            # is the log on the blacklist?
            text = formatter.format(record)
            blacklisted = False
            for string in self.blacklist:
                if string in text:
                    blacklisted = True
                    break
            if blacklisted:
                continue

            # actually log
            print(text)
            if stream is None:  # no more work if we're not using AWS
                continue

            # No, seriously, you have to do this :(
            if sequence_token is None:
                result = aws_logger.put_log_events(
                        logGroupName=group,
                        logStreamName=stream,
                        logEvents=[
                            {
                                'timestamp': int(record.created * 1000),
                                'message': text
                            }
                        ]  # you can't pass None as the sequence token
                )
            else:
                result = aws_logger.put_log_events(
                        logGroupName=group,
                        logStreamName=stream,
                        logEvents=[
                            {
                                'timestamp': int(record.created * 1000),
                                'message': text
                            }
                        ],
                        sequenceToken=sequence_token
                )
            sequence_token = result['nextSequenceToken']




# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""Various utilities"""

import json
import cbor
import requests
import boto3
from botocore.exceptions import ClientError

dd_result_cache = None
dd_have_result = False


def dynamic_data_or_none():
    """Returns the instance identity document on AWS or None"""
    global dd_result_cache, dd_have_result
    if not dd_have_result:
        try:
            dynamic_data_text = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document',
                                             timeout=0.5).text
            dd_result_cache = json.loads(dynamic_data_text)
        except requests.exceptions.ConnectionError:
            dd_result_cache = None
        dd_have_result = True
    return dd_result_cache


def boto_client(type, dynamic_data):
    """Create a boto client. Type is efs, ssm etc."""
    session = boto3.session.Session(region_name=dynamic_data['region'])
    try:
        return session.client(type)
    except ClientError:
        raise RuntimeError("There was a problem creating a client for '%s'\n" +
                           "Please ensure you are running under a profile with the correct permissions.")



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""A drop in replacement for boto3.session.client"""


import json
import socketserver
import _thread
import logging
import os
from . import dynamic_data_or_none, boto_client
from botocore.exceptions import ClientError


class KeyValue:
    fname = None

    def __init__(self, non_aws_filename="kvstore", port=1026, *, noserver=False):
        # Self.ssm gets created if we are under AWS
        KeyValue.fname = non_aws_filename
        self.dynamic_data = dynamic_data_or_none()
        self.ssm = None
        self.server = None
        if self.dynamic_data is not None:
            self.ssm = boto_client('ssm', self.dynamic_data)
        else:
            if '/' in non_aws_filename:  # at least *a* directory
                last_slash = non_aws_filename.rfind('/')
                os.makedirs(non_aws_filename[:last_slash], exist_ok=True)
            if not noserver:
                self.server = socketserver.UDPServer(('0.0.0.0', port), KeyValue._UDPHandler)
                _thread.start_new_thread(socketserver.UDPServer.serve_forever, (self.server,))

    def stop(self):
        if self.server is not None:
            self.server.shutdown()

    # non pep8 names are to retain compatibility with the AWS calls
    def put_parameter(self, Name, Description, Type, Value, Overwrite):
        if self.ssm:
            self.ssm.put_parameter(Name=Name, Description=Description, Type=Type, Value=Value, Overwrite=Overwrite)
        else:
            # write into a file just called "kvstore". create if not there
            values = KeyValue._values()
            if Name in values.keys() and not Overwrite:
                raise ValueError("Tried to overwrite a kv parameter with Overwrite=False")
            values[Name] = Value
            json_string = json.dumps(values, indent=2)
            if len(json_string) > 65536:
                raise RuntimeError("Cannot write parameter, KV storage total allocation is 64K")
            with open(KeyValue.fname, "w") as f:
                f.write(json_string)

    # ditto, retaining compatibility
    def get_parameter(self, Name):
        if self.ssm:
            return self.ssm.get_parameter(Name=Name)
        else:
            values = KeyValue._values()
            try:
                return {'Parameter': {'Value': values[Name]}}
            except KeyError:
                raise ClientError(operation_name='get_parameter',
                                  error_response={'ResponseMetadata': {'MaxAttemptsReached': True}})

    @staticmethod
    def _values():
        try:
            with open(KeyValue.fname, "r") as f:
                return json.loads(f.read())
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            return {}

    class _UDPHandler(socketserver.BaseRequestHandler):
        def handle(self):
            skt = self.request[1]
            try:
                with open(KeyValue.fname, "r") as f:
                    skt.sendto(f.read().encode(), self.client_address)
                logging.debug("Served kv store to: " + str(self.client_address))
            except FileNotFoundError:
                return skt.sendto("{}", self.client_address)



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


from setuptools import setup

setup(name='awsornot',
      version='0.7.1',
      author='David Preece',
      author_email='davep@polymath.tech',
      url='https://polymath.tech',
      license='BSD',
      packages=['awsornot'],
      install_requires=['requests', 'boto3'],
      description='Classes for logging and key/value that work identically whether running on AWS or not',
      long_description="Classes for logging and client/server key value storage that do the same thing " +
                       "regardless of whether or not the code is running on AWS. The logger creates CloudWatch " +
                       "logs asynchronously or logs to stdout; and the KV store replicates put_parameter and " +
                       "get_parameter from the SSM API.",
      keywords='AWS EC2 logging log logger CloudWatch SSM KV Key/Value',
      classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: BSD License',
            'Topic :: System :: Boot',
            'Topic :: System :: Boot :: Init',
            'Topic :: System :: Clustering',
            'Topic :: System :: Distributed Computing',
            'Topic :: System :: Software Distribution',
            'Topic :: System :: Systems Administration',
            'Topic :: Utilities',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6'
      ]
      )
